import os
import cv2
import numpy as np
import onnxruntime
from typing import List, Dict, Optional
from datetime import datetime
import argparse
from tqdm import tqdm

class UAVDetector:
    def __init__(self, model_path: str, prob_threshold: float = 0.5, max_bbox_ratio: float = 0.05):
        # 初始化ONNX会话
        so = onnxruntime.SessionOptions()
        so.log_severity_level = 3
        self.session = onnxruntime.InferenceSession(
            model_path, 
            so,
            providers=['CUDAExecutionProvider', 'CPUExecutionProvider']
        )
        
        # 模型参数设置
        self.input_size = (640, 640)  # width, height
        self.mean_ = np.array([0.485, 0.456, 0.406], dtype=np.float32)
        self.std_ = np.array([0.229, 0.224, 0.225], dtype=np.float32)
        self.confThreshold = prob_threshold
        self.max_bbox_ratio = max_bbox_ratio
        
        # 获取输入输出名称
        self.input_name = self.session.get_inputs()[0].name
        self.output_name = self.session.get_outputs()[0].name

    def preprocess(self, srcimg: np.ndarray) -> tuple:
        """图像预处理"""
        img = cv2.cvtColor(srcimg, cv2.COLOR_BGR2RGB)
        img = cv2.resize(img, self.input_size, interpolation=cv2.INTER_LINEAR)
        img = img.astype(np.float32)
        img = img / 255.
        img -= self.mean_[None, None, :]
        img /= self.std_[None, None, :]
        img = np.transpose(img, [2, 0, 1])
        scale_factor = np.array([1., 1.], dtype=np.float32)
        return img, scale_factor

    def nms(self, boxes: np.ndarray, scores: np.ndarray, conf_threshold: float, iou_threshold: float) -> List[int]:
        """非极大值抑制"""
        x1 = boxes[:, 0]
        y1 = boxes[:, 1]
        x2 = boxes[:, 2]
        y2 = boxes[:, 3]
        areas = (x2 - x1 + 1) * (y2 - y1 + 1)
        order = scores.argsort()[0][::-1]
        keep = []
        
        while order.size > 0:
            i = order[0]
            keep.append(i)
            xx1 = np.maximum(x1[i], x1[order[1:]])
            yy1 = np.maximum(y1[i], y1[order[1:]])
            xx2 = np.minimum(x2[i], x2[order[1:]])
            yy2 = np.minimum(y2[i], y2[order[1:]])
            w = np.maximum(0.0, xx2 - xx1 + 1)
            h = np.maximum(0.0, yy2 - yy1 + 1)
            inter = w * h
            ovr = inter / (areas[i] + areas[order[1:]] - inter)
            inds = np.where(ovr <= iou_threshold)[0]
            order = order[inds + 1]
        return keep

    def detect(self, srcimg: np.ndarray) -> tuple:
        """执行检测"""
        # 预处理
        img, scale_factor = self.preprocess(srcimg)
        inputs = {'image': img[None, :, :, :], 'scale_factor': scale_factor[None, :]}
        ort_inputs = {i.name: inputs[i.name] for i in self.session.get_inputs() if i.name in inputs}
        
        # 推理
        output = self.session.run(None, ort_inputs)
        bbox, confidence = output
        
        # 过滤低置信度检测
        keep_idx = (confidence[:, 0] > self.confThreshold)
        bbox = bbox[keep_idx, :]
        confidence = confidence[:, keep_idx]

        # 坐标转换
        ratioh = srcimg.shape[0] / self.input_size[1]
        ratiow = srcimg.shape[1] / self.input_size[0]
        bbox[:, 0] *= ratiow
        bbox[:, 1] *= ratioh
        bbox[:, 2] *= ratiow
        bbox[:, 3] *= ratioh

        # NMS处理
        keep = self.nms(bbox, confidence, self.confThreshold, 0.4)

        # 后处理
        outres = []
        img_out = srcimg.copy()
        for idx in keep:
            xmin, ymin, xmax, ymax = bbox[idx].astype(int)
            
            # 检查检测框面积比例
            bbox_area = (xmax - xmin) * (ymax - ymin)
            image_area = srcimg.shape[0] * srcimg.shape[1]
            if bbox_area / image_area > self.max_bbox_ratio:
                continue
                
            score = confidence[0][idx]
            outres.append([xmin, ymin, xmax, ymax, score])
            
            # 绘制检测框和标签
            cv2.rectangle(img_out, (xmin, ymin), (xmax, ymax), (0, 0, 255), thickness=2)
            cv2.putText(img_out, f"UAV: {score:.2f}", (xmin, ymin - 10),
                        cv2.FONT_HERSHEY_SIMPLEX, 0.4, (0, 255, 0), thickness=1)
            
        return img_out, outres

def process_image(detector: UAVDetector, image_path: str, output_dir: str) -> int:
    """处理单张图像"""
    srcimg = cv2.imread(image_path)
    if srcimg is None:
        print(f"无法读取图像: {image_path}")
        return 0
        
    processed_img, detections = detector.detect(srcimg)
    
    # 保存结果
    if detections or args.save_empty:
        filename = os.path.basename(image_path)
        output_path = os.path.join(output_dir, filename)
        cv2.imwrite(output_path, processed_img)
        
    return len(detections)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='UAV Detection using ONNX model')
    parser.add_argument('--model', type=str, required=True, help='ONNX模型路径')
    parser.add_argument('--input', type=str, required=True, help='输入图像路径或目录')
    parser.add_argument('--output', type=str, default='output', help='输出目录')
    parser.add_argument('--threshold', type=float, default=0.5, help='检测置信度阈值')
    parser.add_argument('--max-bbox-ratio', type=float, default=0.05, help='检测框最大面积比例阈值')
    parser.add_argument('--save-empty', action='store_true', help='是否保存未检测到目标的图片')
    args = parser.parse_args()

    # 创建输出目录
    os.makedirs(args.output, exist_ok=True)

    # 初始化检测器
    detector = UAVDetector(
        model_path=args.model,
        prob_threshold=args.threshold,
        max_bbox_ratio=args.max_bbox_ratio
    )

    # 处理输入
    if os.path.isdir(args.input):
        total_detections = 0
        image_files = []
        for root, _, files in os.walk(args.input):
            image_files.extend([os.path.join(root, f) for f in files if f.lower().endswith(('.jpg', '.png'))])
            
        for img_file in tqdm(image_files, desc='Processing images'):
            total_detections += process_image(detector, img_file, args.output)
            
        print(f'批量处理完成！共检测到 {total_detections} 个目标')
    else:
        detections = process_image(detector, args.input, args.output)
        print(f'处理完成！检测到 {detections} 个目标') 